// File Name: ResMngr.cpp
// Author: Jesse Towner
// Date: Sept. 14, 1999
//
// Description: Source file for the Resource File Management Library Demo.
//	Contains the implement of the CResFile class, along with various helper
//   functions.

///////////////////////////////////////////////////////////////////////////////
//	INCLUDES

#include <windows.h>
#include <fstream.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>

#include "ResMngr.h"

///////////////////////////////////////////////////////////////////////////////
//	PROTOTYPES

// Lump handling functions for raw data lumps.
static BOOL __stdcall resLoadRawLump(fstream *, LPRESFILELUMP);
static BOOL __stdcall resSaveRawLump(fstream *, LPRESFILELUMP);
static VOID __stdcall resUnloadRawLump(LPRESFILELUMP);

///////////////////////////////////////////////////////////////////////////////
//	FUNCTIONS

//- resLoadRawLump ------------------------------------------------------------
// Description: Loads in the data for a raw data lump.
// Parameters:  lpStream - pointer to the file stream object.
//              lpLump   - pointer to the lump node.
// Returns:     TRUE if successful, FALSE otherwise.
//-----------------------------------------------------------------------------
static BOOL __stdcall resLoadRawLump(fstream *lpStream, LPRESFILELUMP lpLump)
{
	// First, allocate enough memory to the load the data into.
	if ((lpLump->lpData = LPVOID(new CHAR[lpLump->dwSize])) == NULL)
		return FALSE;

	// Next, load in the data.
	lpStream->read(PCHAR(lpLump->lpData), lpLump->dwSize);
	if (lpStream->fail())
	{
		delete [] PCHAR(lpLump->lpData);
		lpLump->lpData = NULL;
		return FALSE;
	}

	// It worked, so return true.
	return TRUE;
}

//- resSaveRawLump ------------------------------------------------------------
// Description: Saves the data of a raw data lump to a file.
// Parameters:  lpStream - pointer to the file stream object.
//              lpLump   - pointer to the lump node.
// Returns:     TRUE if successful, FALSE otherwise.
//-----------------------------------------------------------------------------
static BOOL __stdcall resSaveRawLump(fstream *lpStream, LPRESFILELUMP lpLump)
{
	// Write the data to the file.
	lpStream->write(PCHAR(lpLump->lpData), lpLump->dwSize);
	if (lpStream->fail())
		return FALSE;

	// It worked, so return true!
	return TRUE;
}

//- resUnloadRawLump ----------------------------------------------------------
// Description: Unloads the data of a raw data lump.
// Parameters:  lpLump   - pointer to the lump node.
//-----------------------------------------------------------------------------
static VOID __stdcall resUnloadRawLump(LPRESFILELUMP lpLump)
{
	// Delete the memory.
	delete [] PCHAR(lpLump->lpData);
	lpLump->lpData = NULL;
}

///////////////////////////////////////////////////////////////////////////////
//	CLASS IMPLEMENTATION

BOOL CResFile::m_bHandlerActive = FALSE;
CResFile::LPRESFILELUMPHANDLER CResFile::m_lpLumpHandlerList = NULL;

//- CResFile::GetFileSize -----------------------------------------------------
// Description: Gets the file size of the active file.
// Returns:     Size of the file in bytes, -1 on error.
//-----------------------------------------------------------------------------
LONG CResFile::GetFileSize(VOID)
{
	// Store the current position.
	fstream::pos_type p = m_fStream.tellp();
	if (m_fStream.fail())
		return -1;

	// Seek to the end of the file.
	m_fStream.seekp(0, ios_base::end);
	if (m_fStream.fail())
		return -1;

	// Get the file size.
	LONG lSize = LONG(m_fStream.tellp());
	BOOL bFail = BOOL(m_fStream.fail());

	// Seek back to the original position and return the size.
	m_fStream.seekp(p, ios_base::beg);
	return ((bFail) ? -1 : lSize);
}

//- CResFile::GetLumpHandler --------------------------------------------------
// Description: Returns the pointer to the matching lump handler for a lump
//              type. If no match is found, it returns the raw data lump handler.
// Parameters:  dwType - value for the type of lump that our lump handler should
//                       handle.
// Returns:     Pointer to the correct lump handler.
//-----------------------------------------------------------------------------
CResFile::LPRESFILELUMPHANDLER CResFile::GetLumpHandler(DWORD dwType)
{
	// Search through the lump handler list looking for a matching lump handler.
	LPRESFILELUMPHANDLER lpLumpHandler = m_lpLumpHandlerList, lpRawLumpHandler;
	while (lpLumpHandler)
	{
		// if this is the raw data lump handler, cache a pointer to it for
		// possible future reference.
		if (lpLumpHandler->dwType == RESFILE_RAWLUMP)
			lpRawLumpHandler = lpLumpHandler;

		// Check to see if this is a match, if so, return a pointer to the
		// lump handler.
		if (lpLumpHandler->dwType == dwType)
			return lpLumpHandler;
	}

	// No matching lump was found, so default to the raw data lump handler.
	return lpRawLumpHandler;
}

//- CResFile::LoadLumpList ----------------------------------------------------
// Description: Private method for loading in the lump description table from
//              a resource file into a lump information linked list.
// Returns:     TRUE if successful, FALSE otherwise.
//-----------------------------------------------------------------------------
BOOL CResFile::LoadLumpList(VOID)
{
	// Make sure a file is open for read or modification mode.
	if ((!m_fStream.is_open()) || (m_cFileMode[0] != 'r'))
		return FALSE;

	// Seek to the start of the lump list in the resource file.
	m_fStream.seekp(m_resFileHeader.dwLumpListOffset, ios_base::beg);
	if (m_fStream.fail())
		return FALSE;

	// Read in each node into the linked list.
	LPRESFILELUMP *lplpLump = &m_lpLumpList;
	BOOL bError = FALSE;

	for (DWORD i=0; i<m_resFileHeader.dwNumLumps; i++)
	{
		// Load in the lump information.
		RESFILELUMPINFO lumpInfo;
		m_fStream.read(PCHAR(&lumpInfo), sizeof(lumpInfo));
		if (m_fStream.fail())
		{
			bError = TRUE;
			break;
		}

		// Allocate memory for the node.
		if ((*lplpLump = new RESFILELUMP) == NULL)
		{
			bError = TRUE;
			break;
		}

		// Set up the node.
		(*lplpLump)->dwSize = lumpInfo.dwSize;
		(*lplpLump)->dwOffset = lumpInfo.dwOffset;
		(*lplpLump)->dwType = lumpInfo.dwType;
		(*lplpLump)->bNoFree = FALSE;
		(*lplpLump)->lpData = NULL;
		(*lplpLump)->lplpData = NULL;
		(*lplpLump)->lpNextLump = NULL;

		// Allocate memory for the lump name string.
		if (((*lplpLump)->lpName = new CHAR[lumpInfo.byNumChar + 1]) == NULL)
		{
			bError = TRUE;
			break;
		}

		// Load in the name.
		m_fStream.read((*lplpLump)->lpName, lumpInfo.byNumChar);
		if (m_fStream.fail())
		{
			bError = TRUE;
			break;
		}

		// Set the terminating NULL character for name string.
		(*lplpLump)->lpName[lumpInfo.byNumChar] = NULL;
          lplpLump = &(*lplpLump)->lpNextLump;
	}

	// Check to see if there were any errors.
	if (bError == TRUE)
	{
		// There was an error, unload what was loaded in.
		UnloadLumpList();
		return FALSE;
	}

	// No errors, so return true.
	return TRUE;
}

//- CResFile::SaveLumpList ----------------------------------------------------
// Description: Outputs the lump list to the resource file. Only saves the lump
//              information, not the actual data. The actual data is saved
//              by messaging the method CResFile::SaveLumps().
// Returns:     TRUE if successful, FALSE otherwise.
//-----------------------------------------------------------------------------
BOOL CResFile::SaveLumpList(VOID)
{
	// Make sure a file is open for write or modification mode.
	if ((!m_fSaveStream.is_open()) || ((m_cFileMode[0] != 'w') && (m_cFileMode[1] != '+')))
		return FALSE;

	// Go through the lump linked list, writing out the info for each node in
	// the list.
	LPRESFILELUMP lpLumpList = m_lpLumpList;
	while (lpLumpList)
	{
		// Set up the lump info structure.
		RESFILELUMPINFO lumpInfo;
		lumpInfo.dwSize = lpLumpList->dwSize;
		lumpInfo.dwOffset = lpLumpList->dwOffset;
		lumpInfo.dwType = lpLumpList->dwType;
		lumpInfo.byNumChar = BYTE(strlen(lpLumpList->lpName));

		// Write the lump info structure to the file.
		m_fSaveStream.write(PCHAR(&lumpInfo), sizeof(lumpInfo));
		if (m_fSaveStream.fail())
			return FALSE;

		// Write out the name.
		m_fSaveStream.write(PCHAR(lpLumpList->lpName), lumpInfo.byNumChar);
		if (m_fSaveStream.fail())
			return FALSE;

		// Traverse to the next lump on the list.
		lpLumpList = lpLumpList->lpNextLump;
	}

	// No errors, so return TRUE.
	return TRUE;
}

//- CResFile::UnloadLumpList --------------------------------------------------
// Description: Deallocates the resource file manager's lump list. Used when
//              the file is closed.
//-----------------------------------------------------------------------------
VOID CResFile::UnloadLumpList(VOID)
{
	// Deallocate the entire lump list.
	LPRESFILELUMP *lplpLumpList = &m_lpLumpList;
	while (*lplpLumpList)
	{
		// If the data has been loaded in, unload it.
		if (((*lplpLumpList)->lpData != NULL) && ((*lplpLumpList)->bNoFree == FALSE))
		{
			LPRESFILELUMPHANDLER lpHandler = GetLumpHandler((*lplpLumpList)->dwType);
			lpHandler->lpUnload(*lplpLumpList);
			if ((*lplpLumpList)->lplpData)
				*(*lplpLumpList)->lplpData = NULL;
		}

		// Deallocate the lump name.
		if ((*lplpLumpList)->lpName)
			delete [] PCHAR((*lplpLumpList)->lpName);

		// Traverse to the next node and deallocate the current one.
		LPRESFILELUMP lpLump = *lplpLumpList;
		lplpLumpList = &(lpLump->lpNextLump);
		delete lpLump;
	}

     // Set the pointer to NULL.
     m_lpLumpList = NULL;
}

//- CResFile::SaveLumps -------------------------------------------------------
// Description: Saves the data for each lump to the resource file. Called by
//              CResFile::Save() method.
// Returns:     TRUE if successful, FALSE otherwise.
//-----------------------------------------------------------------------------
BOOL CResFile::SaveLumps(VOID)
{
	// Make sure a file is open for write or modification mode.
	if ((!m_fSaveStream.is_open()) || ((m_cFileMode[0] != 'w') && (m_cFileMode[1] != '+')))
		return FALSE;

	// Run through the lump linked list, writing the data for each lump to the
	// save resource file.
	LPRESFILELUMP lpLumpList = m_lpLumpList;
	while (lpLumpList)
	{
		// If the data has been loaded in, save it to the new file.
		if (lpLumpList->lpData)
		{
			// Get the new file offset for the lump.
			lpLumpList->dwOffset = DWORD(m_fSaveStream.tellp());

			// Find the lump handler for this lump and save the lump.
			LPRESFILELUMPHANDLER lpHandler = GetLumpHandler(lpLumpList->dwType);
			if (lpHandler->lpSave(&m_fSaveStream, lpLumpList) == FALSE)
				return FALSE;

			// Get the lump's size.
			lpLumpList->dwSize = DWORD(m_fSaveStream.tellp()) - lpLumpList->dwOffset;
		}
		else
		{
			// Otherwise, copy the lump's data from the old file to the new one.

			// Seek to the lump position in the old file and store the offset
			// of the lump in the new file.
			m_fStream.seekp(lpLumpList->dwOffset, ios_base::beg);
			lpLumpList->dwOffset = DWORD(m_fSaveStream.tellp());

			// Load in the bytes from the old file and save them out in
			// the new file.
			DWORD dwSize = lpLumpList->dwSize;
			CHAR cBuffer[2048];

			while (dwSize != 0)
			{
				if (dwSize >= 2048)
				{
					m_fStream.read(cBuffer, 2048);
					m_fSaveStream.write(cBuffer, 2048);
					dwSize -= 2048;
				}
				else
				{
					m_fStream.read(cBuffer, dwSize);
					m_fSaveStream.write(cBuffer, dwSize);
					dwSize = 0;
				}
			}
		}

		// Traverse to the next lump in the list.
		lpLumpList = lpLumpList->lpNextLump;
	}

	// Return true, as no errors occured.
	return TRUE;
}

#ifdef __BORLANDC__
VOID _USERENTRY CResFile::Shutdown(VOID)
#else
VOID CResFile::Shutdown(VOID)
#endif
{
	// Delete the lump handler list.
	LPRESFILELUMPHANDLER lpLumpHandler = m_lpLumpHandlerList, lpTemp;
	while (lpLumpHandler)
	{
		lpTemp = lpLumpHandler->lpNextHandler;
		delete lpLumpHandler;
		lpLumpHandler = lpTemp;
	}

	// Set the base node to NULL.
	m_lpLumpHandlerList = NULL;
}

//- CResFile::Constructor -----------------------------------------------------
// Description: Default constructor of this class - clears out data structures.
//-----------------------------------------------------------------------------
CResFile::CResFile()
{
     // Clear out the object.
	memset(&m_resFileHeader, 0, sizeof(m_resFileHeader));
     m_lpLumpList = NULL;
     memset(&m_cFileName, 0, sizeof(m_cFileName));
	memset(&m_cFileMode, 0, sizeof(m_cFileMode));

	// Register the raw lump handling routines.
	if (m_bHandlerActive == FALSE)
	{
		RegisterLumpHandler(RESFILE_RAWLUMP, resLoadRawLump, resSaveRawLump, resUnloadRawLump);

		// Register the shutdown function.
		atexit(CResFile::Shutdown);
		m_bHandlerActive = TRUE;
	}
}

//- CResFile::Open ------------------------------------------------------------
// Description: Opens a resource file either for reading from, writing to, or
//              modification. Read mode opens up a resource file for read only
//              operations, and in this mode, the resource file can not be
//              modified. Write mode creates a new resource file (or overwrites
//              any existing file with the same name) and allows the programmer
//              to add lumps to the file. Modification mode opens up an existing
//              file (or creates one if one doesn't exist) and allows the
//              programmer to add more lumps to the file or to load in the lumps.
// Parameters:  lpFileName - name of the resource file (may include path) to open.
//              lpFileMode - string containing the access mode, can be of the
//                           following contents:
//                           	"r"  - Read Mode
//  						"w"  - Write Mode
//  						"r+" - Modification Mode
// Returns:     TRUE if successful, FALSE otherwise.
//-----------------------------------------------------------------------------
BOOL CResFile::Open(
	LPCSTR lpFileName,
	LPCSTR lpFileMode )
{
	// If the file is already open, then return false.
	if (m_fStream.is_open())
          return FALSE;

	// Set up the file name.
	if (lpFileName != m_cFileName)
          strcpy(m_cFileName, lpFileName);

	// Copy the file mode into our local buffer.
	memset(m_cFileMode, 0, sizeof(m_cFileMode));
	strcpy(m_cFileMode, lpFileMode);

	if (stricmp(m_cFileMode, "w") == 0)
	{
		// Open up the file for read/write replace mode.
		m_fStream.open(m_cFileName, ios_base::in | ios_base::out | ios_base::trunc | ios_base::binary);
		if (!m_fStream.is_open())
               return FALSE;

		// Set up the header.
		memset(&m_resFileHeader, 0, sizeof(m_resFileHeader));
		strncpy(m_resFileHeader.cSignature, RESFILE_SIGNATURE, 4);
		m_resFileHeader.byVersionLo = RESFILE_VERLO;
		m_resFileHeader.byVersionHi = RESFILE_VERHI;

		// We're done as far as setting things up for a new resource file,
		// so return true...
		return TRUE;
	}
	else
     {
          // Open the file for read only.
          if (stricmp(m_cFileMode, "r") == 0)
     		m_fStream.open(m_cFileName, ios_base::in | ios_base::binary);
          // Open the file for modification.
          else if (stricmp(m_cFileMode, "r+") == 0)
               m_fStream.open(m_cFileName, ios_base::in | ios_base::out | ios_base::binary);
          else
               return FALSE;

		// Make sure the file was opened properly.
		if (!m_fStream.is_open())
			return FALSE;

		// Load in the header and make sure it's a valid resource file.
          m_fStream.seekp(0, ios_base::beg);
		m_fStream.read(PCHAR(&m_resFileHeader), sizeof(m_resFileHeader));
		if (m_fStream.fail())
		{
			m_fStream.close();
			return FALSE;
		}

		// Check the signature and the checksum.
		if ((strnicmp(m_resFileHeader.cSignature, RESFILE_SIGNATURE, sizeof(m_resFileHeader.cSignature)) != 0) ||
			((DWORD(GetFileSize()) ^ 0xFFFFFFFF) != m_resFileHeader.dwChecksum))
		{
			// Invalid resource file.
			m_fStream.close();
			return FALSE;
		}

		// Make sure the version of the file isn't greater than the library version.
		DWORD dwVersion = (DWORD(m_resFileHeader.byVersionHi) << 8) | DWORD(m_resFileHeader.byVersionLo);
		if (dwVersion > ((RESFILE_VERHI << 8) | RESFILE_VERLO))
		{
			m_fStream.close();
			return FALSE;
		}

		// Load in the lump list (if any).
		if (LoadLumpList() == FALSE)
		{
			m_fStream.close();
			return FALSE;
		}

		// Everything has gone smoothly, so return true.
		return TRUE;
	}
}

//- CResFile::Save ------------------------------------------------------------
// Description: Saves the contents of a resource file in memory to disk. If
//              for some reason you want to save the file with an alternate
//              file name (ie. SaveAs operations), then pass the new file name
//              string pointer in the "lpAltFileName" parameter. Otherwise,
//              you can just pass NULL for this parameter to keep the original
//              file name.
// Parameters:  lpAltFileName - alternate file name.
// Returns:     TRUE if successful, FALSE otherwise.
//-----------------------------------------------------------------------------
BOOL CResFile::Save(LPCSTR lpAltFileName)
{
	// Make sure a file is open for write or modification mode.
	if ((!m_fStream.is_open()) || ((m_cFileMode[0] != 'w') && (m_cFileMode[1] != '+')))
		return FALSE;

	// Open up a temporary resource file to save the new contents to.
	m_fSaveStream.open(RESFILE_TEMPNAME, ios_base::out | ios_base::trunc | ios_base::binary);
	if (!m_fSaveStream.is_open())
		return FALSE;

	// Write out the header. Note that we must later write it out again after
	// it's fields have been updated, but we're merely doing this so that the
	// lumps are aligned properly.
	m_fSaveStream.write(PCHAR(&m_resFileHeader), sizeof(m_resFileHeader));
	if (m_fSaveStream.fail())
		goto fail;

	// Save all of the lump data to the resource file.
	if (SaveLumps() == FALSE)
		goto fail;

	// Get the offset of the lump list and store it in the header.
	m_resFileHeader.dwLumpListOffset = DWORD(m_fSaveStream.tellp());
	if (m_fSaveStream.fail())
		goto fail;

	// Save the lump list to the resource file.
	if (SaveLumpList() == FALSE)
		goto fail;

	// Get the file size and create the checksum.
	m_resFileHeader.dwChecksum = 0xFFFFFFFF ^ DWORD(m_fSaveStream.tellp());
	if (m_fSaveStream.fail())
		goto fail;

	// Seek back to the start of the file and write out the new updated header.
	m_fSaveStream.seekp(0, ios_base::beg);
	if (m_fSaveStream.fail())
		goto fail;
	m_fSaveStream.write(PCHAR(&m_resFileHeader), sizeof(m_resFileHeader));
	if (m_fSaveStream.fail())
		goto fail;

	// Close the files, and delete the old file if we're overwriting it and
	// rename the temporary file to our original or alternate name.
	m_fSaveStream.close();
	Close();

	if (lpAltFileName)
	{
		// Aquire the full path name to the file using the Win32 API function.
		LPSTR lpFilePart;
		if (GetFullPathName(lpAltFileName, MAX_PATH, m_cFileName, &lpFilePart) != MAX_PATH)
			return FALSE;
	}
	else
	{
		// Delete the old file.
		remove(m_cFileName);
	}

	// After renaming the temporary file, load it back in.
	rename(RESFILE_TEMPNAME, m_cFileName);
	return Open(m_cFileName, "r+");

	// If this method failed, close the save file and delete it.
 fail:
	m_fSaveStream.close();
	remove(RESFILE_TEMPNAME);
	return FALSE;
}

//- CResFile::Close -----------------------------------------------------------
// Description: Closes the resource file if one iscurrently open.
//-----------------------------------------------------------------------------
VOID CResFile::Close(VOID)
{
	// If the file isn't open, return.
	if (!m_fStream.is_open())
		return;

	// Unload the lumps and the lump list.
	UnloadLumpList();

	// Close the file.
	m_fStream.close();
}

//- CResFile::RegisterLumpHandler ---------------------------------------------
// Description: Registers a lump handler with the CResFile class. A lump handler
//              consists of three functions, one each for loading, saving, and
//              unloading a lump of a particular type. You must also pass the
//              value which will used to identify wether lumps should be
//              handled by the lump handler or not - each value must be unique.
// Parameters:  dwType   - the type of lumps this handler will handle.
//              lpLoad   - pointer to the loading function.
//              lpSave   - pointer to the saving function.
//              lpUnload - pointer tothe unloading function.
// Returns:     TRUE if successful, FALSE otherwise.
//----------------------------------------------------------------------------
BOOL CResFile::RegisterLumpHandler(
	DWORD dwType,
	BOOL __stdcall (* lpLoad)(fstream *, LPRESFILELUMP),
	BOOL __stdcall (* lpSave)(fstream *, LPRESFILELUMP),
	VOID __stdcall (* lpUnload)(LPRESFILELUMP) )
{
	// Traverse to the end of the lump handler list where we will add a new
	// lump handler node.
	LPRESFILELUMPHANDLER *lplpHandler = &m_lpLumpHandlerList;
	while (*lplpHandler)
	{
		// Make sure that a handler for this lump type hasn't already been
		// registered...
		if ((*lplpHandler)->dwType == dwType)
			return FALSE;

		// traverse to the next node on the list.
		lplpHandler = &((*lplpHandler)->lpNextHandler);
	}

	// Allocate memory for the new lump handler and attach it to the list.
	if ((*lplpHandler = new RESFILELUMPHANDLER) == NULL)
		return FALSE;

	// Set up the lump handler node.
	(*lplpHandler)->dwType = dwType;
	(*lplpHandler)->lpLoad = lpLoad;
	(*lplpHandler)->lpSave = lpSave;
	(*lplpHandler)->lpUnload = lpUnload;
	(*lplpHandler)->lpNextHandler = NULL;

	// Return true now that we have successfully registered a new lump handler.
	return TRUE;
}

//- CResFile::RemoveLumpHandler -----------------------------------------------
// Description: Removes a lump handler previously added by a message to the
//              CResFile::RegisterLumpHandler() method.
// Parameters:  dwType - the lump type of the handler to remove.
// Returns:     TRUE if successful, FALSE otherwise.
//----------------------------------------------------------------------------
BOOL CResFile::RemoveLumpHandler(DWORD dwType)
{
	// Search the lump handler list for a lump type match. If one is found,
	// then remove the corresponding lump handler.
	LPRESFILELUMPHANDLER *lplpHandler = &m_lpLumpHandlerList;
	while (*lplpHandler)
	{
		// Check to see if we have a match; if there is a match, remove the
		// lump handler.
		if ((*lplpHandler)->dwType == dwType)
		{
			LPRESFILELUMPHANDLER lpHandler = *lplpHandler;
			*lplpHandler = (*lplpHandler)->lpNextHandler;
			delete lpHandler;       // De-allocate the memory for the handler.
			return TRUE;
		}

		// Traverse to the next node in the list.
		lplpHandler = &((*lplpHandler)->lpNextHandler);
	}

	// No match for the lump type was found, so return false.
	return FALSE;
}

//- CResFile::LumpExists ------------------------------------------------------
// Description: Checks to see if a lump, with a particular name, exists.
// Parameters:  lpName - string of the lump to check for existance.
// Returns:     TRUE if it exists, FALSE if it doesn't.
//-----------------------------------------------------------------------------
BOOL CResFile::LumpExists(LPCSTR lpName)
{
	// Search through the lump list, looking for a matching lump.
	LPRESFILELUMP lpLumpList = m_lpLumpList;
	while (lpLumpList)
	{
		// If we have a match, return true.
		if (stricmp(lpLumpList->lpName, lpName) == 0)
			return TRUE;

		// Traverse to the next lump in the list.
		lpLumpList = lpLumpList->lpNextLump;
	}

	// No matching lump was found, return false.
	return FALSE;
}

//- CResFile::CreateLump ------------------------------------------------------
// Description: Creates a new lump and adds it to the active resource file.
//              Note that it will notbe saved with the file unless the
//              CResFile::Save() method is messaged.
// Parameters:  lpName - Name of the lump.
//              dwType - Particular type of the lump (used for loading/saving)
//              lpData - Pointer to the data or data structure that will be
//                       stored in the lump.
//              dwSize - Size of the data in bytes (used only for RAW data lumps).
//              bFree  - Set this to TRUE if you want the data (pointed to by
//                       lpData) to be deallocated when the file is closed.
// Returns:     TRUE is successful, FALSE otherwise.
//-----------------------------------------------------------------------------
BOOL CResFile::CreateLump(
	LPCSTR lpName,
	DWORD dwType,
	LPVOID lpData,
	DWORD dwSize,
	BOOL bFree )
{
	// First, make sure the file is open for write or modification mode.
	if ((!m_fStream.is_open()) || ((m_cFileMode[0] != 'w') && (m_cFileMode[1] != '+')))
		return FALSE;

	// Make sure the lump name is under 256 characters.
	if (strlen(lpName) >= 256)
		return FALSE;

	// Traverse to the end of the lump list. Make sure not to create any lumps
	// with the same names.
	LPRESFILELUMP *lplpLumpList = &m_lpLumpList;
	while (*lplpLumpList)
	{
		if (stricmp((*lplpLumpList)->lpName, lpName) == 0)
			return FALSE;

		lplpLumpList = &((*lplpLumpList)->lpNextLump);
	}

	// Allocate memory for the lump node.
	if ((*lplpLumpList = new RESFILELUMP) == NULL)
		return FALSE;

	// Set up the lump node.
	(*lplpLumpList)->dwSize = dwSize;
	(*lplpLumpList)->dwOffset = 0;
	(*lplpLumpList)->dwType = dwType;
	(*lplpLumpList)->bNoFree = !bFree;
	(*lplpLumpList)->lpData = lpData;
	(*lplpLumpList)->lpNextLump = NULL;

	// Allocate memory for the name string.
	if (((*lplpLumpList)->lpName = new CHAR[strlen(lpName) + 1]) == NULL)
	{
		delete *lplpLumpList;
		*lplpLumpList = NULL;
		return FALSE;
	}

	// Copy the name into the lump node's local name string.
	strcpy((*lplpLumpList)->lpName, lpName);

     // Increment the number of lumps in the file.
     m_resFileHeader.dwNumLumps++;

	// Everything went nice and slick, so return true.
	return TRUE;
}

//- CResFile::DeleteLump ------------------------------------------------------
// Description: Removes or deletes a particular lump, designated by 'lpName',
//              from the active resource file.
// Parameters:  lpName - name of the lump to remove from the file.
// Returns:     TRUE if successful, FALSE otherwise.
//-----------------------------------------------------------------------------
BOOL CResFile::DeleteLump(LPCSTR lpName)
{
	// Traverse the lump linked list, looking for a lump matching the name
	// located in the 'lpName' string.
	LPRESFILELUMP *lplpLumpList = &m_lpLumpList;
	while (*lplpLumpList)
	{
		// If the name's are the same, we found a match, so remove the lump.
		if (stricmp((*lplpLumpList)->lpName, lpName) == 0)
		{
			// Unload the lump's data if we have to.
			if ((*lplpLumpList)->lpData)
			{
				LPRESFILELUMPHANDLER lpLumpHandler = GetLumpHandler((*lplpLumpList)->dwType);
				lpLumpHandler->lpUnload(*lplpLumpList);
				if ((*lplpLumpList)->lplpData)
					*(*lplpLumpList)->lplpData = NULL;
			}

			// Delete the lump's name.
			if ((*lplpLumpList)->lpName)
				delete [] PCHAR((*lplpLumpList)->lpName);

			// Delete the lump from the list.
			LPRESFILELUMP lpLump = *lplpLumpList;
			*lplpLumpList = lpLump->lpNextLump;
			delete lpLump;

               // Decrement the number of lumps in the file.
               m_resFileHeader.dwNumLumps--;

			// Return true, now that we've deleted the lump.
			return TRUE;
		}

		// Traverse to the next lump on the list.
		lplpLumpList = &((*lplpLumpList)->lpNextLump);
	}

	// Damn, there was no lump matching the name, so return false.
	return FALSE;
}

//- CResFile::LoadLump --------------------------------------------------------
// Description: Loads in a lump from a resource file. Depending upon it's
//              type, it will used the specially designed routine for loading
//              it in. However, if such a routine doesn't exist, it will
//              default the lump as raw data.
// Parameters:  lpName   - name of the lump to load in.
//              lplpData - pointer to the location where the address of the data
//                         or data structure will be stored.
// Returns:     TRUE if successful, FALSE otherwise.
//-----------------------------------------------------------------------------
BOOL CResFile::LoadLump(LPCSTR lpName, LPVOID *lplpData)
{
	// Make sure the file is open.
	if (!m_fStream.is_open())
		return FALSE;

	// Search through the lump list for a macthing lump name.
	LPRESFILELUMP lpLumpList = m_lpLumpList;
	while (lpLumpList)
	{
		// Check if a match was found.
		if (stricmp(lpLumpList->lpName, lpName) == 0)
		{
			// If the data hasn't been loaded in, load it in!
			if (!lpLumpList->lpData)
			{
				// Seek to the lump's position in the file.
				m_fStream.seekp(lpLumpList->dwOffset, ios_base::beg);
				if (m_fStream.fail())
					return FALSE;

				// Get the lump handler for this lump type and load in the lump.
				LPRESFILELUMPHANDLER lpHandler = GetLumpHandler(lpLumpList->dwType);
				if (lpHandler->lpLoad(&m_fStream, lpLumpList) == FALSE)
					return FALSE;
			}

			// Set up the address variable and return true!
			lpLumpList->lplpData = lplpData;
			*lplpData = lpLumpList->lpData;
			return TRUE;
		}

		// Traverse to the next node in the list.
		lpLumpList = lpLumpList->lpNextLump;
	}

	// No lump matching the name was found, so return FALSE :(
	return FALSE;
}

//- CResFile::UnloadLump ------------------------------------------------------
// Description: Unloads a lump from memory that was previously loaded in from
//              a resource file using the CResFile::LoadLump() method.
// Parameters:  lpName - name of the lump to unload from memory.
// Returns:     TRUE if successful, FALSE otherwise.
//-----------------------------------------------------------------------------
BOOL CResFile::UnloadLump(LPCSTR lpName)
{
	// Search through the lump list for a lump with a matching name.
	LPRESFILELUMP lpLumpList = m_lpLumpList;
	while (lpLumpList)
	{
		// Check to see if this is our match...
		if (stricmp(lpLumpList->lpName, lpName) == 0)
		{
			// Unload the lump's data.
			LPRESFILELUMPHANDLER lpHandler = GetLumpHandler(lpLumpList->dwType);
			lpHandler->lpUnload(lpLumpList);
			if (lpLumpList->lplpData)
				lpLumpList->lplpData = NULL;

			// Return true now that we have unloaded the lump's data.
			return TRUE;
		}

		// Traverse to the next lump in the list.
		lpLumpList = lpLumpList->lpNextLump;
	}

	// No lumps were found matching the name, so return false.
	return FALSE;
}

//- CResFile::Destructor ------------------------------------------------------
// Description: Deallocates any memory and closes the resource file if it
//              is still open.
//-----------------------------------------------------------------------------
CResFile::~CResFile()
{
	// Close the file if it's open.
	Close();
}

///////////////////////////////////////////////////////////////////////////////

// end of file //
