// File Name: ResMngr.h
// Author: Jesse Towner
// Date: Sept. 14, 1999
//
// Description: Header file for the Resource File Management Library Demo.
// 	Contains the class definition for the CResFile resource file management
//   class. The CResFile class demonstrates one method of managing resource
//   files and allows a programmer to easily use such files.

#ifndef ResMngrH
#define ResMngrH

///////////////////////////////////////////////////////////////////////////////
//	DEFINES

// Temporary file name used for saving a resource file.
#define RESFILE_TEMPNAME "rf$$$$$.res"

// Header defines.
#define RESFILE_SIGNATURE "RESF"
#define RESFILE_VERHI     0x01
#define RESFILE_VERLO     0x00

// Resource file lump types.
#define RESFILE_RAWLUMP   0

///////////////////////////////////////////////////////////////////////////////
//	STRUCTURES

// Pack structures on byte alignment.
#pragma pack(push,1)

// Structure defining the header of our resource files.
typedef struct tagRESFILEHEADER
{
	CHAR cSignature[4];               // 4-character identification value.
	BYTE byVersionLo;                 // Minor version.
	BYTE byVersionHi;                 // Major version.
	WORD wFlags;                      // Special flags for the resource file.
	DWORD dwNumLumps;                 // Number of lumps in the resource file.
	DWORD dwChecksum;                 // Checksum (0FFFFFFFFh ^ filesize).
	DWORD dwLumpListOffset;           // Offset of the lump info list.
} RESFILEHEADER, *LPRESFILEHEADER;

// Structure of each element in the lump info list. Holds information which
// can be used to retrieve the contents of a lump.
typedef struct tagRESFILELUMPINFO
{
	DWORD dwSize;                     // Size of the lump, in bytes.
	DWORD dwOffset;                   // Offset in the file of the lump.
	DWORD dwType;                     // The type of lump.
	BYTE byNumChar;                   // Length in characters of the lump's name.

	// ... The lump's name string goes here, but it can be of variable length;
	// therefore, we manually load it in to memory pointed to by 'lpName' in
	// the RESFILELUMP structure.

} RESFILELUMPINFO, *LPRESFILELUMPINFO;

// Restore structure packing alignment value.
#pragma pack(pop)

// Structure used to keep track of a lump in memory.
typedef struct tagRESFILELUMP
{
	DWORD dwSize;                     // Size of the lump, in bytes.
	DWORD dwOffset;                   // Offset in the file of the lump.
	DWORD dwType;                     // The type of lump.
	BOOL bNoFree;                     // Temporary flags.
	LPSTR lpName;                     // Pointer to the name of the lump.
	LPVOID lpData;                    // Pointer to the data of the lump.
	LPVOID *lplpData;                 // Pointer to the data address varaible.
	tagRESFILELUMP *lpNextLump;       // Pointer to the next lump in the list.
} RESFILELUMP, *LPRESFILELUMP;

///////////////////////////////////////////////////////////////////////////////
//	CLASS DEFINITIONS

// The actual Resource File Management class definition. The objects of the
// CResFile class can be used to create, load, or save resource files.
class CResFile
{
 protected:
	// Internal structure used to keep track of a lump handler.
	typedef struct tagRESFILELUMPHANDLER
	{
		DWORD dwType;
		BOOL __stdcall (* lpLoad)(fstream *, LPRESFILELUMP);
		BOOL __stdcall (* lpSave)(fstream *, LPRESFILELUMP);
		VOID __stdcall (* lpUnload)(LPRESFILELUMP);
		tagRESFILELUMPHANDLER *lpNextHandler;
	} RESFILELUMPHANDLER, *LPRESFILELUMPHANDLER;

	static BOOL m_bHandlerActive;
	static LPRESFILELUMPHANDLER m_lpLumpHandlerList;

	RESFILEHEADER m_resFileHeader;   // Header structure for the resource file.
	LPRESFILELUMP m_lpLumpList;      // Pointer to the root of the lump linked list.

	CHAR m_cFileName[MAX_PATH];      // Local storage of the resource file's name.
	CHAR m_cFileMode[4];             // Current file access mode.
	fstream m_fStream;               // File stream object.
	fstream m_fSaveStream;           // Temporary file stream object for save operations.

 // Private internal methods.
	LONG GetFileSize(VOID);
	LPRESFILELUMPHANDLER GetLumpHandler(DWORD);
	BOOL LoadLumpList(VOID);
	BOOL SaveLumpList(VOID);
	VOID UnloadLumpList(VOID);
	BOOL SaveLumps(VOID);

 #ifdef __BORLANDC__
	static VOID _USERENTRY Shutdown(VOID);
 #else
	static VOID Shutdown(VOID);
 #endif

 public:

 //- Constructor --------------------------------------------------------------
 // Description: Default constructor of this class - clears out data structures.
 //----------------------------------------------------------------------------
	CResFile();

 //- Constructor --------------------------------------------------------------
 // Description: See the CResFile::Open() method description.
 //----------------------------------------------------------------------------
	CResFile(LPCSTR lpFileName, LPCSTR lpFileMode)
		{ this->Open(lpFileName, lpFileMode); }

 //- Open ---------------------------------------------------------------------
 // Description: Opens a resource file either for reading from, writing to, or
 //              modification. Read mode opens up a resource file for read only
 //              operations, and in this mode, the resource file can not be
 //              modified. Write mode creates a new resource file (or overwrites
 //              any existing file with the same name) and allows the programmer
 //              to add lumps to the file. Modification mode opens up an existing
 //              file (or creates one if one doesn't exist) and allows the
 //              programmer to add more lumps to the file or to load in the lumps.
 // Parameters:  lpFileName - name of the resource file (may include path) to open.
 //              lpFileMode - string containing the access mode, can be of the
 //                           following contents:
 //                           	"r"  - Read Mode
 //							"w"  - Write Mode
 //							"r+" - Modification Mode
 // Returns:     TRUE if successful, FALSE otherwise.
 //----------------------------------------------------------------------------
	BOOL Open(LPCSTR lpFileName, LPCSTR lpFileMode);

 //- Save ---------------------------------------------------------------------
 // Description: Saves the contents of a resource file in memory to disk. If
 //              for some reason you want to save the file with an alternate
 //              file name (ie. SaveAs operations), then pass the new file name
 //              string pointer in the "lpAltFileName" parameter. Otherwise,
 //              you can just pass NULL for this parameter to keep the original
 //              file name.
 // Parameters:  lpAltFileName - alternate file name.
 // Returns:     TRUE if successful, FALSE otherwise.
 //----------------------------------------------------------------------------
	BOOL Save(LPCSTR lpAltFileName);

 //- Close --------------------------------------------------------------------
 // Description: Closes the resource file if one iscurrently open.
 //----------------------------------------------------------------------------
	VOID Close(VOID);

 //- RegisterLumpHandler ------------------------------------------------------
 // Description: Registers a lump handler with the CResFile class. A lump handler
 //              consists of three functions, one each for loading, saving, and
 //              unloading a lump of a particular type. You must also pass the
 //              value which will used to identify wether lumps should be
 //              handled by the lump handler or not - each value must be unique.
 // Parameters:  dwType   - the type of lumps this handler will handle.
 //              lpLoad   - pointer to the loading function.
 //              lpSave   - pointer to the saving function.
 //              lpUnload - pointer tothe unloading function.
 // Returns:     TRUE if successful, FALSE otherwise.
 //----------------------------------------------------------------------------
	static BOOL RegisterLumpHandler(DWORD dwType, BOOL __stdcall (* lpLoad)(fstream *, LPRESFILELUMP), BOOL __stdcall (* lpSave)(fstream *, LPRESFILELUMP), VOID __stdcall (* lpUnload)(LPRESFILELUMP));

 //- RemoveLumpHandler --------------------------------------------------------
 // Description: Removes a lump handler previously added by a message to the
 //              CResFile::RegisterLumpHandler() method.
 // Parameters:  dwType - the lump type of the handler to remove.
 // Returns:     TRUE if successful, FALSE otherwise.
 //----------------------------------------------------------------------------
	static BOOL RemoveLumpHandler(DWORD dwType);

 //- LumpExists ---------------------------------------------------------------
 // Description: Checks to see if a lump, with a particular name, exists.
 // Parameters:  lpName - string of the lump to check for existance.
 // Returns:     TRUE if it exists, FALSE if it doesn't.
 //----------------------------------------------------------------------------
	BOOL LumpExists(LPCSTR lpName);

 //- CreateLump ---------------------------------------------------------------
 // Description: Creates a new lump and adds it to the active resource file.
 //              Note that it will notbe saved with the file unless the
 //              CResFile::Save() method is messaged.
 // Parameters:  lpName - Name of the lump.
 //              dwType - Particular type of the lump (used for loading/saving)
 //              lpData - Pointer to the data or data structure that will be
 //                       stored in the lump.
 //              dwSize - Size of the data in bytes (used only for RAW data lumps).
 //              bFree  - Set this to TRUE if you want the data (pointed to by
 //                       lpData) to be deallocated when the file is closed.
 // Returns:     TRUE is successful, FALSE otherwise.
 //----------------------------------------------------------------------------
	BOOL CreateLump(LPCSTR lpName, DWORD dwType, LPVOID lpData, DWORD dwSize, BOOL bFree=FALSE);

 //- DeleteLump ---------------------------------------------------------------
 // Description: Removes or deletes a particular lump, designated by 'lpName',
 //              from the active resource file.
 // Parameters:  lpName - name of the lump to remove from the file.
 // Returns:     TRUE if successful, FALSE otherwise.
 //----------------------------------------------------------------------------
	BOOL DeleteLump(LPCSTR lpName);

 //- LoadLump -----------------------------------------------------------------
 // Description: Loads in a lump from a resource file. Depending upon it's
 //              type, it will used the specially designed routine for loading
 //              it in. However, if such a routine doesn't exist, it will
 //              default the lump as raw data.
 // Parameters:  lpName   - name of the lump to load in.
 //              lplpData - pointer to the location where the address of the data
 //                         or data structure will be stored.
 // Returns:     TRUE if successful, FALSE otherwise.
 //----------------------------------------------------------------------------
	BOOL LoadLump(LPCSTR lpName, LPVOID *lplpData);

 //- UnloadLump ---------------------------------------------------------------
 // Description: Unloads a lump from memory that was previously loaded in from
 //              a resource file using the CResFile::LoadLump() method.
 // Parameters:  lpName - name of the lump to unload from memory.
 // Returns:     TRUE if successful, FALSE otherwise.
 //----------------------------------------------------------------------------
	BOOL UnloadLump(LPCSTR lpName);

 //- Destructor ---------------------------------------------------------------
 // Description: Deallocates any memory and closes the resource file if it
 //              is still open.
 //----------------------------------------------------------------------------
	~CResFile();
};

///////////////////////////////////////////////////////////////////////////////

#endif

// end of file //
